/**********************************************************************/
/* SimCell: Cell/B.E. Processor Simulator        Arch Lab. TOKYO TECH */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <elf.h>
#include <sys/stat.h>
#include <sys/time.h>

/** Type Definition ***************************************************/
/**********************************************************************/
typedef unsigned int       uint;
typedef signed long long   llint;
typedef unsigned long long ullint;

typedef float              flot_032t;
typedef double             flot_064t;
typedef signed char        sint_008t;
typedef signed short       sint_016t;
typedef signed int         sint_032t;
typedef signed long long   sint_064t;
typedef unsigned char      uint_008t;
typedef unsigned short     uint_016t;
typedef unsigned int       uint_032t;
typedef unsigned long long uint_064t;
typedef struct {
    uint_064t high;
    uint_064t low;
} uint_128t;

/** Constant Definition ***********************************************/
/**********************************************************************/
#define L_NAME "SimCell: Cell/B.E. Simulator"
#define L_VER  "Version 0.8.1 2008-05-02"

#define PROCESSOR_FREQ 3200.0f      // processor frequency in MHz
#define AVERAGE_IPC 1.0f            // processor average IPC

#define MAX_SPES 512
#define MAX_CYCLE_DEF 1000000000    // limit of simulation cycle
#define VERBOSE_INT_DEF  1000000    // verbose interval cycle
#define LOGMODE_INT_DEF  1000000    // log-mode interval cycle
#define LOGMODE_FILE "log.dat"      // memory access log written to

// EIB network latency
#define MEM_RDDELAY_DEF 500
#define MEM_WRDELAY_DEF 300
#define MEM_WIDTH_DEF 16
#define C2C_DELAY_DEF 300
#define C2C_WIDTH_DEF 8

/**********************************************************************/
enum {
    NREG = 128,                     // Number of Resisters
    REG_LENGTH = 16,                // Length of a Register (Byte)
    LS_SIZE = (256 * 1024),         // Local Store Size (Byte)
    CH_NUM = 31,                    // Number of Channels
    MFC_MAX_ENTRY = 16,
    EIB_MAX_ENTRY = 4,
    EIB_WAIT_ENTRY = MAX_SPES,
    MAX_DEBUG_MODE = 3,
    SPE_RUN = 0,
    SPE_STOP = 1,
    SPE_STALL = 2,
    SPE_HALT = 3,
    MEM_READ = 0,
    MEM_WRITE = 1,
    MEM_DMA = 2,
    MEM_INST_READ = 3,
    MEM_ACCESS_KIND = 4,
};

/** Class and Function Definition *************************************/
/**********************************************************************/
class Spe;
class MainMemory;
class Mmu;
class Eib;
class SpuInstinfo;

/**********************************************************************/
class Chip {
 private:
    bool checkendian();

 public:
    int spe_num;
    ullint cycle;
    ullint max_cycle;
    ullint verb_cycle;
    ullint log_cycle;
    int debug_mode;
    bool log_mode;
    bool force_mode;
    bool big_endian;
    int mem_rddelay;
    int mem_wrdelay;
    int mem_width;
    int c2c_delay;
    int c2c_width;
    int running_core;
    int ***logdata;
    int logpage;
    FILE *logfile;
    char *mem_imagefile;

    Spe *spe[MAX_SPES];
    MainMemory *mem;
    Mmu *mmu;
    Eib *eib;
    SpuInstinfo *iinfo;

    Chip();
    void setlogmode();
    void initlogmode();
    void setlogdata(int, int, int);
    void writelogdata();
    void finilogmode();
};

/** SPU Global Perpose Register ***************************************/
#define NSLOT(data_type) (int)(REG_LENGTH / sizeof(data_type))

typedef union {
    uint_008t i008t[NSLOT(uint_008t)];
    uint_016t i016t[NSLOT(uint_016t)];
    uint_032t i032t[NSLOT(uint_032t)];
    uint_064t i064t[NSLOT(uint_064t)];
    flot_032t f032t[NSLOT(uint_032t)];
    flot_064t f064t[NSLOT(uint_064t)];
} regdata_t;

/**********************************************************************/
class Register {
  private:
    Chip *chip;
    regdata_t reg;

  public:
    Register(Chip *);

    void print();

    uint_008t get008t(int);
    uint_016t get016t(int);
    uint_032t get032t(int);
    uint_064t get064t(int);
    uint_128t get128t();
    flot_032t getf32t(int);
    flot_064t getf64t(int);

    void put008t(int, uint_008t);
    void put016t(int, uint_016t);
    void put032t(int, uint_032t);
    void put064t(int, uint_064t);
    void put128t(uint_128t);
    void putf32t(int, flot_032t);
    void putf64t(int, flot_064t);
};

/** FP Status and Control Register ************************************/
class Fpscr {
 private:
    Chip *chip;
    regdata_t reg;
    regdata_t enable;

 public:
    Fpscr(Chip *);

    uint_128t read();
    void write(uint_128t);
    void bitset(int);
    void twobitwrite(int, int);
};

/** Channel and Channel Interface *************************************/
class Channel {
private:
    uint_032t *message, ch_count;
    int head, tail;
    const uint_032t size;

public:
    const int blocking_mode, channel_type;

    Channel(int, int, int);
    ~Channel();
    
    bool empty();
    bool full();
    uint_032t capacity();
    uint_032t occupancy();
    void inccnt();
    void deccnt();
    uint_032t read();
    void write(uint_032t);
};

/**********************************************************************/
class ChannelInterface {
 public:
    Channel *ch[CH_NUM];
    
    ChannelInterface();
    ~ChannelInterface();
    
    bool spureadable(int);
    bool spuwritable(int);
    int spureadch(int, uint_032t*);
    int spuwritech(int, uint_032t);
    uint_032t spureadchcnt(int);
    void mfcreadch(int, uint_032t*);
    void mfcwritech(int, uint_032t);
};

/** Local Store *******************************************************/
class LocalStore {
 private:
    uint_032t *mem;

 public:
    LocalStore(ullint);
    ~LocalStore();

    void read1b(uint_032t, uint_008t*);
    void read4b(uint_032t, uint_032t*);
    void readnb(uint_032t, int, uint_032t[]);
    void write1b(uint_032t, uint_008t);
    void write4b(uint_032t, uint_032t);
    void writenb(uint_032t, int, uint_032t[]);
    void print();
};

/** Architecture State ************************************************/
class SpeArchstate {
 private:
    Chip *chip;
 public:
    uint_032t pc;
    uint lslr;
    int state;
    Register **reg;
    Fpscr *fpscr;
    int stop_reason;

    SpeArchstate(Chip*);
    ~SpeArchstate();
};

/** Simulator State ***************************************************/
class SpeSimstate {
 public:
    ullint instcnt;
    int stall_reason;

    SpeSimstate();
};

/** SPU Instructions **************************************************/
enum {
    IT_NONE = 0x0,
    IT_W_RT = 0x1,
    IT_R_RA = 0x2,
    IT_R_RB = 0x4,
    IT_R_RC = 0x8,
    IT_R_RT = 0x10,
    IT_R_I18 = 0x20,
    IT_R_I16 = 0x40,
    IT_R_I10 = 0x80,
    IT_R_I8 = 0x100,
    IT_R_I7 = 0x200,
    IT_R_CA = 0x400,
    IT_R_RO1 = 0x800,
    IT_R_RO2 = 0x1000,
    NOADDR = 0xffffffff,
    IINFO_OP_NUM = 0x800,
    IINFO_ID_NUM = 0x1000,
    PIPE_EVEN = 0,
    PIPE_ODD = 1,
};

/**********************************************************************/
typedef struct {
    uint type;
    uint pipe;
    uint latency;
    uint stall;
    char *name;
} spuinstattr_t;

/**********************************************************************/
class SpuInstinfo {
 private:
    int op_to_id[IINFO_OP_NUM];
    spuinstattr_t id_to_attr[IINFO_ID_NUM];

    int predecode(int);
    void setattr();
 public:
    SpuInstinfo(Chip *);
    int getinstid(int);
    spuinstattr_t getinstattr(int);
    char *getopname(int);
};

/**********************************************************************/
class SpuInst {
 private:
    Chip *chip;
    Spe *spe;
    SpeArchstate *as;
    SpuInstinfo *iinfo;

 public:
    uint_032t ir, pc, nextpc, targ;
    uint op, imm, ca, ro;
    uint type, pipe, latency, stall;
    uint depend;
    uint ra, rb, rc, rt;
    Register *rra, *rrb, *rrc, *rrt;
    int nextstate;

    SpuInst(Chip*, Spe*);
    ~SpuInst();
    char *getopname();
    void exprinthead();
    void exprinttail();
    int isload();
    int isstore();
    int isbranch();
    int ishintbranch();
    void decode();
    void clear();
};

/** Synergistic Processor Unit ****************************************/
class Spu {
  private:
    Chip *chip;
    Spe *spe;
    ChannelInterface *chi;
    SpeArchstate *as;
    SpeSimstate *ss;

    void load(uint_032t, Register *);
    void store(uint_032t, Register *);

    int checkstate();
    void fetch(SpuInst*);
    void decode(SpuInst*);
    void executeodd(SpuInst*);
    void executeeven(SpuInst*);
    void execute(SpuInst*);
    void lsaccess(SpuInst*);
    void writeback(SpuInst*);
    void setnextpc(SpuInst*);
    
  public:
    SpuInst *inst;

    Spu(Chip*, Spe*);
    ~Spu();
    void printregall();
    void message();
    void start();
    void step();

};

/** Memory Flow Controller ********************************************/
typedef struct{
    uint_032t logical;
    uint_032t physical;
    int unitid;
} address_t;

/**********************************************************************/
typedef struct {
    address_t lsa;
    address_t ea;
    uint_032t size;
    uint_032t tagid;
    uint_032t classid;
    uint_032t cmd;
} dmacmd_t;

/**********************************************************************/
typedef struct {
    int delay;
    int state;
} dmacmdtag_t;

/**********************************************************************/
class Mfc {
 private:
    Chip *chip;
    Mmu *mmu;
    Spe *spe;
    ChannelInterface *chi;
    
    dmacmd_t mfcspucmdq[MFC_MAX_ENTRY];
    int head, tail, entry;
    dmacmdtag_t cmdtag[MFC_MAX_ENTRY];

    uint_032t tagmask;
    uint_032t tagupdate;
    uint_032t tagstat;
    uint_032t decrementer;

    dmacmd_t getcmdfromchi();
    int enq_spucmd(dmacmd_t, dmacmdtag_t);
    int deq_spucmd();
    bool ismemread(dmacmd_t);
    bool ismemwrite(dmacmd_t);
    int getdelay(dmacmd_t);
    void makedmacmd();
    void checkchannel();
    void checkdma();
    uint_032t maketagstat();
    void updatetagstate();
    void decrementerstep();
    
public:
    Mfc(Chip*, Spe*);

    void step();
};

/** Synergistic Processor Element *************************************/
class Spe {
 private:
    Chip *chip;

 public:
    Spu *spu;
    Mfc *mfc;
    LocalStore *ls;
    ChannelInterface *chi;
    SpeArchstate *as;
    SpeSimstate *ss;

    int unitid;

    Spe(Chip *, int);
    ~Spe();

    void message(ullint);
};

/** Main Memory *******************************************************/
enum {
    MAIN_MEM_SIZE = 0x100000000ull, // Main Memory size
    BLOCK_SIZE = (4 * 1024),        // Page size
    BLOCK_TABLE_SIZE = (MAIN_MEM_SIZE / BLOCK_SIZE),
};

/**********************************************************************/
class MainMemory {
 private:
    uint_032t *block_table[BLOCK_TABLE_SIZE];
    
    uint_032t *allocblock(uint_032t);

 public:
    MainMemory(Chip *);
    ~MainMemory();
    
    void read1b(uint_032t, uint_008t*);
    void read2b(uint_032t, uint_016t*);
    void read4b(uint_032t, uint_032t*);
    void read8b(uint_032t, uint_064t*);
    void readnb(uint_032t, int, uint_032t[]);
    void write1b(uint_032t, uint_008t);
    void write2b(uint_032t, uint_016t);
    void write4b(uint_032t, uint_032t);
    void write8b(uint_032t, uint_064t);
    void writenb(uint_032t, int, uint_032t[]);
    void print();
};

/** Memory Management Unit ********************************************/
typedef struct{
    uint_032t addr;
    uint_032t id;
}mapdata_t;

/**********************************************************************/
class Mmu {
 private:
    mapdata_t mmutable[BLOCK_TABLE_SIZE];

 public:
    Mmu(Chip *);

    void transaddr(address_t*);
    int mapping(uint_032t, int, uint_032t);
    void mapls(uint_032t, int);
};

/** Element Interconnect Bus ******************************************/
typedef struct{
    dmacmd_t *cmd;
    dmacmdtag_t *tag;
} eibcmd_t;

/**********************************************************************/
class Eib {
private:
    Chip *chip;
    Spe *spe[MAX_SPES];
    MainMemory *mem;
    uint_032t *transdata_ptr;

    eibcmd_t activecmd[EIB_MAX_ENTRY];
    int valid[EIB_MAX_ENTRY];
    int entry_count;
    eibcmd_t waitcmdq[EIB_WAIT_ENTRY];
    int head, tail;

    int deq_waitcmd(eibcmd_t*);
    bool emptyactivecmd();
    bool fullactivecmd();
    void activatecmd();
    void checkcmd(dmacmd_t*);
    void transdata(dmacmd_t*);
    void comstep();

 public:    
    Eib(Chip*);
    ~Eib();

    int enq_waitcmd(dmacmd_t*, dmacmdtag_t*);
    void step();
};

/** ELF Loader ********************************************************/
#define EM_SPU 23

/**********************************************************************/
typedef struct {
    uint id;
    uint addr;
    uint data;
} memdata;

/**********************************************************************/
typedef struct {
    uint addr;
    uint size;
    char *name;
} functable;

/**********************************************************************/
class Loaderbe {
 private:
    uint getword(uint);
    uint gethword(uint);
    int checkendian();
    uint big_endian;
 public:
    memdata *md;
    int archtype;
    int filetype;
    int needintp;
    uint entry;

    functable *ft;

    Loaderbe();
    ~Loaderbe();
    int loadelf(char *);
};

/** functions *********************************************************/
// in spe.cc
uint_032t exts32(uint_032t, int);
void spefunccall(Spe *);
// in etc.cc
ullint gettime();
void readmemimage(Chip *, uint_032t *);
int issettingfile(char *);
void readsetting(Chip *, char *, char **);
void cyclelog(Chip *);
ullint atoi_postfix(const char *);

/** SPU instruction ID Definition *************************************/
/**********************************************************************/
// Memory - Load/Store Instructions
#define LQD______ 0x001
#define LQX______ 0x002
#define LQA______ 0x003
#define LQR______ 0x004
#define STQD_____ 0x005
#define STQX_____ 0x006
#define STQA_____ 0x007
#define STQR_____ 0x008
#define CBD______ 0x009
#define CBX______ 0x00a
#define CHD______ 0x00b
#define CHX______ 0x00c
#define CWD______ 0x00d
#define CWX______ 0x00e
#define CDD______ 0x00f
#define CDX______ 0x010

// Constant-Formation Instructions
#define ILH______ 0x101
#define ILHU_____ 0x102
#define IL_______ 0x103
#define ILA______ 0x104
#define IOHL_____ 0x105
#define FSMBI____ 0x106

// Integer and Logical Instructions
#define AH_______ 0x201
#define AHI______ 0x202
#define A________ 0x203
#define AI_______ 0x204
#define SFH______ 0x205
#define SFHI_____ 0x206
#define SF_______ 0x207
#define SFI______ 0x208
#define ADDX_____ 0x209
#define CG_______ 0x20a
#define CGX______ 0x20b
#define SFX______ 0x20c
#define BG_______ 0x20d
#define BGX______ 0x20e
#define MPY______ 0x20f
#define MPYU_____ 0x210
#define MPYI_____ 0x211
#define MPYUI____ 0x212
#define MPYA_____ 0x213
#define MPYH_____ 0x214
#define MPYS_____ 0x215
#define MPYHH____ 0x216
#define MPYHHA___ 0x217
#define MPYHHU___ 0x218
#define MPYHHAU__ 0x219
#define CLZ______ 0x21a
#define CNTB_____ 0x21b
#define FSMB_____ 0x21c
#define FSMH_____ 0x21d
#define FSM______ 0x21e
#define GBB______ 0x21f
#define GBH______ 0x220
#define GB_______ 0x221
#define AVGB_____ 0x222
#define ABSDB____ 0x223
#define SUMB_____ 0x224
#define XSBH_____ 0x225
#define XSHW_____ 0x226
#define XSWD_____ 0x227
#define AND______ 0x228
#define ANDC_____ 0x229
#define ANDBI____ 0x22a
#define ANDHI____ 0x22b
#define ANDI_____ 0x22c
#define OR_______ 0x22d
#define ORC______ 0x22e
#define ORBI_____ 0x22f
#define ORHI_____ 0x230
#define ORI______ 0x231
#define ORX______ 0x232
#define XOR______ 0x233
#define XORBI____ 0x234
#define XORHI____ 0x235
#define XORI_____ 0x236
#define NAND_____ 0x237
#define NOR______ 0x238
#define EQV______ 0x239
#define SELB_____ 0x23a
#define SHUFB____ 0x23b

// Shift and Rotate Instructions
#define SHLH_____ 0x301
#define SHLHI____ 0x302
#define SHL______ 0x303
#define SHLI_____ 0x304
#define SHLQBI___ 0x305
#define SHLQBII__ 0x306
#define SHLQBY___ 0x307
#define SHLQBYI__ 0x308
#define SHLQBYBI_ 0x309
#define ROTH_____ 0x30a
#define ROTHI____ 0x30b
#define ROT______ 0x30c
#define ROTI_____ 0x30d
#define ROTQBY___ 0x30e
#define ROTQBYI__ 0x30f
#define ROTQBYBI_ 0x310
#define ROTQBI___ 0x311
#define ROTQBII__ 0x312
#define ROTHM____ 0x313
#define ROTHMI___ 0x314
#define ROTM_____ 0x315
#define ROTMI____ 0x316
#define ROTQMBY__ 0x317
#define ROTQMBYI_ 0x318
#define ROTQMBYBI 0x319
#define ROTQMBI__ 0x31a
#define ROTQMBII_ 0x31b
#define ROTMAH___ 0x31c
#define ROTMAHI__ 0x31d
#define ROTMA____ 0x31e
#define ROTMAI___ 0x31f

// Compare, Branch, and Halt Instructions
#define HEQ______ 0x401
#define HEQI_____ 0x402
#define HGT______ 0x403
#define HGTI_____ 0x404
#define HLGT_____ 0x405
#define HLGTI____ 0x406
#define CEQB_____ 0x407
#define CEQBI____ 0x408
#define CEQH_____ 0x409
#define CEQHI____ 0x40a
#define CEQ______ 0x40b
#define CEQI_____ 0x40c
#define CGTB_____ 0x40d
#define CGTBI____ 0x40e
#define CGTH_____ 0x40f
#define CGTHI____ 0x410
#define CGT______ 0x411
#define CGTI_____ 0x412
#define CLGTB____ 0x413
#define CLGTBI___ 0x414
#define CLGTH____ 0x415
#define CLGTHI___ 0x416
#define CLGT_____ 0x417
#define CLGTI____ 0x418
#define BR_______ 0x419
#define BRA______ 0x41a
#define BRSL_____ 0x41b
#define BRASL____ 0x41c
#define BI_______ 0x41d
#define IRET_____ 0x41e
#define BISLED___ 0x41f
#define BISL_____ 0x420
#define BRNZ_____ 0x421
#define BRZ______ 0x422
#define BRHNZ____ 0x423
#define BRHZ_____ 0x424
#define BIZ______ 0x425
#define BINZ_____ 0x426
#define BIHZ_____ 0x427
#define BIHNZ____ 0x428

// Hint-for-Branch Instructions
#define HBR______ 0x501
#define HBRA_____ 0x502
#define HBRR_____ 0x503

// Floating-Point Instructions
#define FA_______ 0x601
#define DFA______ 0x602
#define FS_______ 0x603
#define DFS______ 0x604
#define FM_______ 0x605
#define DFM______ 0x606
#define FMA______ 0x607
#define DFMA_____ 0x608
#define FNMS_____ 0x609
#define DFNMS____ 0x60a
#define FMS______ 0x60b
#define DFMS_____ 0x60c
#define DFNMA____ 0x60d
#define FREST____ 0x60e
#define FRSQEST__ 0x60f
#define FI_______ 0x610
#define CSFLT____ 0x611
#define CFLTS____ 0x612
#define CUFLT____ 0x613
#define CFLTU____ 0x614
#define FRDS_____ 0x615
#define FESD_____ 0x616
#define DFCEQ____ 0x617
#define DFCMEQ___ 0x618
#define DFCGT____ 0x619
#define DFCMGT___ 0x61a
#define DFTSV____ 0x61b
#define FCEQ_____ 0x61c
#define FCMEQ____ 0x61d
#define FCGT_____ 0x61e
#define FCMGT____ 0x61f
#define FSCRWR___ 0x620
#define FSCRRD___ 0x621

// Control Instructions
#define STOP_____ 0x701
#define STOPD____ 0x702
#define LNOP_____ 0x703
#define NOP______ 0x704
#define SYNC_____ 0x705
#define DSYNC____ 0x706
#define MFSPR____ 0x707
#define MTSPR____ 0x708

// Channel Instructions
#define RDCH_____ 0x801
#define WRCH_____ 0x802
#define RCHCNT___ 0x803

// Undefined (Illegal Instruction)
#define UNDEF____ 0xfff

/**********************************************************************/
