/**********************************************************************/
/*  SimCell: Simulator of Cell/B.E.              Arch Lab. TOKYO TECH */
/**********************************************************************/
#include "define.h"

/**********************************************************************/
Loaderbe::Loaderbe()
{
    md = NULL;
    ft = NULL;
    big_endian = checkendian();
}

/**********************************************************************/
Loaderbe::~Loaderbe()
{
    if (ft != NULL) {
        functable *temp;
        for (temp = ft; temp->name != NULL; temp++) {
            delete[]temp->name;
            temp->name = NULL;
        }
        delete[]ft;
        ft = NULL;
    }
    if (md != NULL) {
        delete[]md;
        md = NULL;
    }
}

/**********************************************************************/
uint Loaderbe::gethword(uint data)
{
    if (big_endian) {
        return data;
    } else {
        return ((data & 0xff) << 8) | ((data & 0xff00) >> 8);
    }
}

/**********************************************************************/
uint Loaderbe::getword(uint data)
{
    if (big_endian) {
        return data;
    } else {
        return ((data & 0xff) << 24) | ((data & 0xff00) << 8) |
            ((data & 0xff0000) >> 8) | ((data & 0xff000000) >> 24);
    }
}

/**********************************************************************/
int Loaderbe::checkendian()
{
    union {
        unsigned short sh;
        unsigned char ch[2];
    } test;
    test.sh = 0xfeff;
    return (test.ch[0] == 0xfe); 
}

/**********************************************************************/
int Loaderbe::loadelf(char *filename)
{
    int fd = 0;
    struct stat sb;
    char elf_magic[6] = { ELFMAG0, ELFMAG1, ELFMAG2, ELFMAG3,
        ELFCLASS32, ELFDATA2MSB
    };
    char *file;
    uint *temp;
    uint fsize;
    uint offset;
    int mdcount = 0;
    uint *mdoff;
    uint i, j, k, imax;
    Elf32_Ehdr *ehdr;
    Elf32_Phdr **phdr;
    Elf32_Shdr *shdr;
    Elf32_Sym *symtab;
    uint symtablen;
    char *shstrtab, *strtab, *funcname;

    // open file
    fd = open(filename, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "## Can't find file : %s\n", filename);
        close(fd);
        return 1;
    }

    fstat(fd, &sb);
    fsize = sb.st_size / sizeof(char);
    file = new char[(fsize + 3) & 0xfffffffc];

    // read file
    i = read(fd, file, fsize);
    close(fd);
    if (i != fsize) {
        fprintf(stderr, "## Can't read file : %s\n", filename);
        delete[]file;
        file = NULL;
        return 1;
    }
    // ELF header
    ehdr = (Elf32_Ehdr *) file;
    for (i = 0; i < sizeof(elf_magic); i++) {
        if (ehdr->e_ident[i] != elf_magic[i]) {
            fprintf(stderr, "## not a big-endian ELF32 : %s\n",
                    filename);
            delete[]file;
            file = NULL;
            return 1;
        }
    }
    filetype = gethword(ehdr->e_type);
    archtype = gethword(ehdr->e_machine);
    entry = getword(ehdr->e_entry);
    needintp = 0;

    // program header
    imax = gethword(ehdr->e_phnum);
    phdr = new Elf32_Phdr *[imax];
    mdoff = new uint[imax];

    offset = getword(ehdr->e_phoff);
    mdcount = 0;
    for (i = 0; i < imax; i++) {
        phdr[i] = (Elf32_Phdr *) (file + offset);
        mdoff[i] = mdcount;
        if (getword(phdr[i]->p_type) == PT_INTERP)
            needintp = 1;
        else if (getword(phdr[i]->p_type) == PT_LOAD)
            mdcount += (getword(phdr[i]->p_memsz) + 3) >> 2;
        offset += gethword(ehdr->e_phentsize);
    }

    // write struct memdata
    md = new memdata[mdcount + 1];
    md[mdcount].id = 0;
    mdcount = 0;

    for (i = 0; i < imax; i++) {
        if (getword(phdr[i]->p_type) == PT_LOAD) {
            k = mdoff[i];
            for (j = 0; j < getword(phdr[i]->p_memsz); j += 4) {
                md[k].id = 4;
                md[k].addr = getword(phdr[i]->p_vaddr) + j;
                md[k].data = 0;
                k++;
            }
            if (getword(phdr[i]->p_memsz) % 4)
                md[k - 1].id = (getword(phdr[i]->p_memsz) % 4);
            temp = (uint *) (file + getword(phdr[i]->p_offset));
            k = mdoff[i];
            for (j = 0; j < getword(phdr[i]->p_filesz); j += 4) {
                md[k].data = getword(*temp);
                k++;
                temp++;
            }
        }
    }

    // section header and symbol table
    imax = gethword(ehdr->e_shnum);
    offset = getword(ehdr->e_shoff);
    symtab = NULL;
    strtab = NULL;
    symtablen = 0;
    shdr = (Elf32_Shdr *) (file + offset + gethword(ehdr->e_shentsize)
                           * gethword(ehdr->e_shstrndx));
    shstrtab = file + getword(shdr->sh_offset);

    for (i = 0; i < imax; i++) {
        shdr = (Elf32_Shdr *) (file + offset);
        if (getword(shdr->sh_type) == SHT_SYMTAB) {
            symtab = (Elf32_Sym *) (file + getword(shdr->sh_offset));
            symtablen = getword(shdr->sh_size) / sizeof(Elf32_Sym);
        }
        if (strcmp(shstrtab + getword(shdr->sh_name), ".strtab") == 0)
            strtab = file + getword(shdr->sh_offset);
        offset += gethword(ehdr->e_shentsize);
    }
    if (symtab != NULL) {
        imax = 0;
        for (i = 0; i < symtablen; i++) {
            if (ELF32_ST_TYPE(symtab[i].st_info) == STT_FUNC)
                imax++;
        }
        ft = new functable[imax + 1];
        j = 0;
        for (i = 0; i < symtablen; i++) {
            if (ELF32_ST_TYPE(symtab[i].st_info) == STT_FUNC) {
                ft[j].addr = getword(symtab[i].st_value);
                ft[j].size = getword(symtab[i].st_size);
                funcname = strtab + getword(symtab[i].st_name);
                k = strlen(funcname);
                ft[j].name = new char[k + 1];
                strncpy(ft[j].name, funcname, k + 1);
                j++;
            }
        }
        ft[imax].name = NULL;
    }
    delete[]mdoff;
    mdoff = NULL;
    delete[]phdr;
    phdr = NULL;
    delete[]file;
    file = NULL;
    return 0;
}

/**********************************************************************/
